#!/usr/bin/env python3
"""Open data file and get spectrum."""
import os
import sys
from pathlib import Path
from argparse import Action
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
from scipy.interpolate import CubicSpline
from maddaq.cmmds.fit_utils import draw_best_fit, fit_gaussian, fit_two_peaks
from maddaq import MadDAQData
from maddaq import ShowProgress

from numpy.polynomial.polynomial import polyfit, polyval
from numpy import interp, ndarray, piecewise


def interp1d(x: ndarray, xp, fp):
    """1D piecewise linear interpolation with linear extrapolation."""
    return piecewise(
        x,
        [x < xp[0], (x >= xp[0]) & (x <= xp[-1]), x > xp[-1]],
        [
            lambda xi: polyval(xi, polyfit(xp[:5], fp[:5], 1)),
            lambda xi: interp(xi, xp, fp),
            lambda xi: polyval(xi, polyfit(xp[-5:], fp[-5:], 1)),
        ],
    )


def draw_hist_and_projection(axh, axp, data, title=None, axis_title=None, x_label=None, y_label=None):
    """Plots a bar histogram and the Y projection.

    Args:
    ----
        axh: axis for the histogram
        axp: axis for the projection
        data: The monitor data
        title: the histogram title
        axis_title: title for the histogram axis
        x_label: label for the histgram X axis
        y_label: label for the Y axis

    Returns
    -------
        mean. std: a tuple with the Y maverage and std

    """
    # Draw the histogram
    y = data
    x = np.linspace(0, len(data)+1, len(data))

    axh.step(x, y)
    if axis_title:
        axh.set_title(axis_title)

    if x_label:
        axh.set_xlabel(x_label)

    if y_label:
        axh.set_ylabel(y_label)

    mean = np.mean(y)
    std = np.std(y)

    # Now the projection
    axp.hist(y, orientation="horizontal")

    return mean, std


def show_pedestals(md):
    """Show module pedestals.

    Args:
    ----
        md: Module object.

    """
    # Get the data and plot
    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(ncols=2, nrows=2, width_ratios=(2, 1))

    # pedestals
    ax = fig.add_subplot(gs[0, 0])
    mped, stped = draw_hist_and_projection(
        ax, fig.add_subplot(gs[0, 1], sharey=ax),
        md.pedestal,
        axis_title="Pedestals",
        x_label="Channel"
    )
    print("Pedestal mean {:.1f} std {:.1f}".format(mped, stped))

    # noise
    ax = fig.add_subplot(gs[1, 0])
    mnoise, stnoise = draw_hist_and_projection(
        ax, fig.add_subplot(gs[1, 1], sharey=ax),
        md.noise,
        axis_title="Noise",
        x_label="Channel"
    )
    print("Noise mean {:.1f} std {:.1f}".format(mnoise, stnoise))


def do_getSpectrum(files, options):
    """Main entry."""

    amplitudes = []
    channels = []

    names = []
    weights = []

    if options.calib:
        _adc, _E = np.loadtxt(options.calib, skiprows=1, delimiter=',', unpack=True)
        cs = CubicSpline(_adc, _E)
        ps = np.polyfit(_adc[:4], _E[:4], 1)
        pf = np.poly1d(ps)

        def get_E(adc):
            """Transform adc into keV"""
            return piecewise(adc,
                             [adc < _adc[0], (adc >= _adc[0]) & (adc < _adc[-1]), adc >= _adc[-1]],
                             [
                                 lambda x: pf(x),
                                 lambda x: cs(x),
                                 np.NaN
                             ],)

    else:
        def get_E(adc):
            """Dummy."""
            return adc

    mx_nchan = 0
    for fnam in files:
        # Check that the file exists
        if not os.path.exists(fnam):
            print("Input file", fnam, "does not exist")
            continue

        i1 = fnam.rfind('/')
        i2 = fnam.rfind('_')
        nam = fnam[i1+1:i2]
        names.append(nam)

        # We open here the file with MadDAQData
        print("\n\n### Opening {}".format(Path(fnam).name))
        maddaq = MadDAQData(fnam)
        maddaq.show_info(True)

        if options.time:
            weights.append(maddaq.get_duration())
        else:
            weights.append(1.0)

        # Find the module
        keys = list(maddaq.modules.keys())
        if options.mid in keys:
            mid = options.mid
        else:
            mid = keys[0]

        print("\n## Looking at data from module {}.".format(mid))
        md = maddaq.modules[mid]
        md.seed_cut = options.seed_cut
        md.neigh_cut = options.neigh_cut
        mx_nchan = max(mx_nchan, md.ntot)

        # Show pedeswtals and noise
        if options.show_ped:
            show_pedestals(md)

        # Create the iterator
        if options.scan_point is not None:
            maddaq_iter = maddaq.create_iterator_at_scanpoint(options.scan_point, [mid])
        elif options.start_time is not None:
            maddaq_iter = maddaq.create_iterator_at_time(options.start_time, [mid])
        elif options.start_evt is not None:
            maddaq_iter = maddaq.create_iterator_at_event(options.start_evt, [mid])
        else:
            maddaq_iter = maddaq.create_module_iterator(md)

        amplitude = []
        chanlist = []
        prg = ShowProgress(maddaq.nevts, width=24)
        prg.start()

        for ievt, evt in enumerate(maddaq_iter):
            data = md.process_event(evt)
            if data is not None:
                for C, E in data:
                    if E > options.thrs and E < options.max_E:
                        amplitude.append(E)
                        chanlist.append(C)

            prg.increase(show=True)

            if ievt > options.nevt:
                break

        amplitudes.append(get_E(amplitude))
        channels.append(chanlist)
        prg.stop()
        print("")

    # Get the normalization factors
    if len(weights) == 0:
        print("No valid file left for analysis. Quitting")
        sys.exit(-1)

    vmax = np.amax(weights)
    weights = [vmax/W for W in weights]

    # Draw the signal
    if options.calib:
        fig, ax = plt.subplots(1, 1)
        xxx = np.linspace(0, 1750, int(1750/2.5))
        yyy = get_E(xxx)
        ax.plot(xxx, yyy)
        ax.grid()
        ax.set_xlabel("ADC counts")
        ax.set_ylabel("Energy (keV)")

    fig, ax = plt.subplots(1, 1)
    i = 0
    ymax = -1e100
    for A, nam in zip(amplitudes, names):
        n, bins, P = ax.hist(A, bins=options.nbin,
                             histtype='stepfilled', alpha=0.25, label=nam)

        if options.time:
            W = weights[i]
            for p in P:
                points = p.get_xy()
                new_points = np.zeros(points.shape)
                ipoint = 0
                for x in points:
                    y = W*x[1]
                    ymax = max(y, ymax)
                    new_points[ipoint, :] = [x[0], y]
                    ipoint += 1

                p.set_xy(new_points)

            i += 1

    if options.time:
        y0, y1 = ax.get_ylim()
        ax.set_ylim(y0, 1.1*ymax)

    # n, bins, *_ = ax.hist(amplitudes, bins=options.nbin, label=names)
    legends = []
    if options.fit:
        for amplitude in amplitudes:
            mean = np.mean(amplitude)
            std = np.std(amplitude)
            if options.two_peaks:
                result, out, legend = fit_two_peaks(mean, std, std, n, bins)

            else:
                result, out, legend = fit_gaussian(n, bins, mean, width=std)

            draw_best_fit(ax, result, bins)
            legends.append(legend)

        ax.legend(legends, loc=1)
    else:
        ax.legend(names, loc=1)

    ax.grid()
    ax.set_title("Signal")
    if options.calib:
        ax.set_xlabel("Energy (keV)")
    else:
        ax.set_xlabel("Charge (ADC)")

    if options.logY:
        ax.set_yscale("log")

    if options.hitmap:
        fig, ax = plt.subplots(1, 1)
        i = 0
        ymax = -1e100
        for A, nam in zip(channels, names):
            n, bins, P = ax.hist(A, bins=int(mx_nchan),
                                 histtype='stepfilled', alpha=0.25, label=nam)
        ax.set_xlabel("channel number")
        ax.set_ylabel("Number  of hits")
        ax.legend()

    plt.show()


class CommaSeparatedListAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        setattr(namespace, self.dest, list(map(int, values.split(','))))


def getSpectrum():
    """Main entry."""

    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--nevt", default=sys.maxsize, type=int, help="max number of events.")
    parser.add_argument("--start-time", dest="start_time", default=None, type=float,
                        help="Event time to start")
    parser.add_argument("--start-evt", dest="start_evt", default=None,
                        action=CommaSeparatedListAction,
                        help="Event number to start showing")
    parser.add_argument("--scan-point", dest="scan_point", type=int, default=None,
                        help="Scan point number to visit")
    parser.add_argument("--nbin", default=50, type=int, help="Number of bins in histogram.")
    parser.add_argument("--logY", default=False, action="store_true", help="Log axis")
    parser.add_argument("--thrs", default=0.0, type=float, help="Min E to show in histogram")
    parser.add_argument("--max_E", default=sys.float_info.max, type=float, help="Max E to show in histogram")
    parser.add_argument("--two_peaks", default=False, action="store_true", help="Min E to show in histogram")
    parser.add_argument("--no-fit", dest="fit", default=True, action="store_false")
    parser.add_argument("--show-ped", action="store_true", default=False, dest="show_ped")
    parser.add_argument("--hitmap", action="store_true", default=False, help="Show hitmap")
    parser.add_argument("--mid", dest="mid", default=-1, type=int, help="The module ID")
    parser.add_argument("--calib", default=None, help="The energy calibration file. X:adc, Y:energy")
    parser.add_argument("--time", default=False, action="store_true",
                        help="Normalize to duration of run when reading various files")
    parser.add_argument("--seed_cut", dest="seed_cut", default=3.5, type=float,
                        help="Threshold of charge for a cluster seed")
    parser.add_argument("--neigh_cut", dest="neigh_cut", default=1.5, type=float,
                        help="Threshold of charge to add a neighbour to a cluster")

    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_getSpectrum(options.files, options)


if __name__ == "__main__":
    getSpectrum()



#!/usr/bin/env python
""" A Timer class
"""
import time


class GTimer(object):
    """ A timer
    """

    def __init__(self, do_start=False):
        self._start = time.time()
        self._running = True
        self._end = self._start
        self._mark = self._start
        if do_start:
            self.start()

    def mark(self):
        return time.time() - self._mark

    def set_mark(self):
        self._mark = time.time()

    def start(self):
        self._running = True
        self._start = time.time()

    def stop(self):
        self._end = time.time()
        self._running = False
        return self._end - self._start

    def reset(self):
        self._start = time.time()

    def __call__(self):
        if self._running:
            return time.time() - self._start
        else:
            return self._end - self._start


if __name__ == "__main__":
    T = GTimer()
    print("Measuring 0.5 seconds")
    T.start()
    time.sleep(0.5)
    print("%.2f" % T.stop())



#!/usr/bin/env python3
"""Python script to show the data in a file."""
import datetime
import os
import sys
from argparse import Action
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
from maddaq import MadDAQData
from maddaq import MadDAQModule
from matplotlib.widgets import Button


def do_show_data(fname, options):
    """The main function."""
    # Check that the file exists
    if not os.path.exists(fname):
        print("Input file", fname, "does not exist")
        return

    # We open here the file with MadDAQData
    maddaq = MadDAQData(fname)
    maddaq.show_info()

    n_mod = len(maddaq.modules)
    fig, ax = plt.subplots(nrows=1, ncols=n_mod)
    dax = {}
    try:
        for a, m in zip(ax, maddaq.modules.values()):
            dax[m.id] = a

    except TypeError:
        m = list(maddaq.modules.values())[0]
        dax[m.id] = ax

    mod_evt = {}
    mod_dtime = {}
    for m in maddaq.modules.values():
        mod_evt[m.id] = 0.0
        mod_dtime[m.id] = 0.0

        m.seed_cut = options.seed_cut
        m.neigh_cut = options.neigh_cut

    plt.subplots_adjust(bottom=0.2)

    if options.scan_point is not None:
        maddaq_iter = maddaq.create_iterator_at_scanpoint(options.scan_point)
    elif options.start_time is not None:
        maddaq_iter = maddaq.create_iterator_at_time(options.start_time)
    elif options.start_evt is not None:
        maddaq_iter = maddaq.create_iterator_at_event(options.start_evt)
    else:
        maddaq_iter = iter(maddaq)

    def next_event(dummy):
        evt = next(maddaq_iter)
        mid = evt.mod_id
        md = maddaq.modules[mid]
        nbin = md.pedestal.shape[0]
        nchan = evt.data.shape[0]
        x = np.arange(0, nbin, 1)

        # Get the indices and build up
        if evt.romode == MadDAQModule.SPARSE_ADJ:
            indx = [i + evt.chan for i in md.adjacents]
            values = np.zeros(nbin)
            try:
                values[indx] = evt.data

            except IndexError:
                print("Bin {} out of bounds for module {}".format(indx, mid))

        elif evt.romode == MadDAQModule.SERIAL:
            values = np.array(evt.data, dtype='d')

        else:
            print("Readout mode not supported:", evt.romode)

        if options.pedestals:
            values -= md.pedestal
            sn = values/md.noise
            cmmd = np.mean(values[np.nonzero(abs(sn) < 5.0)])
            values -= cmmd

        dt = evt.time-mod_dtime[mid]
        mod_dtime[mid] = evt.time
        out = md.process_event(evt)
        clst = "".join(["({:>3d} {:>6.1f}) ".format(int(chan), peak) for chan, peak in out])
        print("Module {}: daq_time {:>10.6f} evt no. {:>6d} evt d_time {:>10.3f} clusters {}".format(mid,
                                                                                                     evt.daq_time/1.e6,
                                                                                                     evt.evtcnt,
                                                                                                     dt*0.025,
                                                                                                     clst))

        ax = dax[mid]
        ax.clear()
        ax.set_title('Module %d' % mid)
        # if options.pedestals:
        #    ax.set_ylim(-250, 2048)
        # else:
        #    ax.set_ylim([0, 4096])

        ax.step(x, values, where='mid')

        plt.draw()

    def loop_events(dummy):
        while options.plot:
            next_event(None)
            plt.draw()
            plt.pause(0.01)

        plt.show()

    axnext = plt.axes([0.81, 0.05, 0.1, 0.075])
    bnext = Button(axnext, 'Next')
    bnext.on_clicked(next_event)

    # axloop = plt.axes([0.4, 0.05, 0.2, 0.075])
    # bloop = Button(axloop, 'Loop/Step')
    # bloop.on_clicked(loop_events)

    axquit = plt.axes([0.1, 0.05, 0.1, 0.075])
    bquit = Button(axquit, 'Quit')
    bquit.on_clicked(sys.exit)

    if options.plot:
        loop_events(None)

    next_event(None)
    plt.show()


class CommaSeparatedListAction(Action):
    """Create a list from the comma sepparated numbers at imput."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual action."""
        setattr(namespace, self.dest, list(map(int, values.split(','))))


def show_data():
    """Define arguments and call main function."""

    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--pedestals",
                        dest="pedestals", action="store_true",
                        help="Remove pedestals on the plot",
                        default=False)
    parser.add_argument("--start-time", dest="start_time", default=None, type=float,
                        help="Event time to start showing")
    parser.add_argument("--start-evt", dest="start_evt", default=None,
                        action=CommaSeparatedListAction,
                        help="Event number to start showing")
    parser.add_argument("--scan-point", dest="scan_point", type=int, default=None,
                        help="Scan point number to visit")
    parser.add_argument("--plot", dest="plot", default=False,
                        action="store_true",
                        help="Draw histograms")
    parser.add_argument("--seed_cut", dest="seed_cut", default=3.5, type=float,
                        help="Threshold of charge for a cluster seed")
    parser.add_argument("--neigh_cut", dest="neigh_cut", default=1.5, type=float,
                        help="Threshold of charge to add a neighbour to a cluster")
    parser.add_argument("--raw-data", dest="raw_data",
                        default=False, action="store_true",
                        help="Show the raw data instead of the 'digested' signal")

    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_show_data(options.files[0], options)


if __name__ == "__main__":
    show_data()


#!/usr/bin/env python3
"""A number of utils to make fits with numpy."""
import matplotlib.pyplot as plt
import numpy as np
from lmfit import Model
from lmfit import Parameters
from lmfit.models import GaussianModel
from scipy.signal import find_peaks
from scipy.stats import moyal as landau

log2 = np.log(2)
s2pi = np.sqrt(2*np.pi)
spi = np.sqrt(np.pi)
s2 = np.sqrt(2.0)
tiny = np.finfo(np.float64).eps


def fLandau(x, norm, mode, width):
    """Approximate the landau distribution with the Moyal distribution.

    Fit model Parameters:
        p[0] = normalization
        p[1] = mode
        p[2] = width

    """
    L = (x-mode)/width
    r = norm*np.exp(-0.5*(L+np.exp(-L)))/2.506628274631001
    return r


def fTwoPeak(x, A1, A2, center, sigma, separation):
    """Two gausians, same width separated by an adjustable amount."""
    std2 = sigma**2
    factor = max(tiny, s2pi*std2)
    factor2 = max(tiny, 2.0*std2)
    shft = separation/2.0
    c1 = center - shft
    c2 = center + shft
    out = (A1/factor)*np.exp(-(x-c1)**2/factor2) + (A2/factor)*np.exp(-(x-c2)**2/factor2)

    return out


def create_multi_peak(peaks):
    """Create a multi gaussan model.

    input is an array of (amplitude, center, width) tuples.
    """
    def create_single_peak(ipeak, center, width=5.0, amplitude=1):
        """Create a single gaussian with initial values as given."""
        pref = "f{0}_".format(ipeak)
        model = GaussianModel(prefix=pref)
        model.set_param_hint(pref+'amplitude', value=amplitude)
        model.set_param_hint(pref+'center', value=center)
        model.set_param_hint(pref+'sigma', value=width)
        return model

    ipeak = 0
    mod = None
    for ampl, center, sigma in peaks:
        this_mod = create_single_peak(ipeak, center, sigma, ampl)
        if mod is None:
            mod = this_mod
        else:
            mod = mod + this_mod

        ipeak += 1

    return mod


def fit_two_peaks(n, bins, center, sigma, separation):
    """Fit 2 gausians with same width.

    Fit Parameters:
        n: histogram bin  contents
        bins: bin limits
        A1, A2: estimated amplitudes of the peaks
        center: estimated center of the reference peak
        sigma: estimated sigma, same for both peaks
        separation: separation between peaks

    Returns
    -------
        result: The fit result object
        out: a (center, sigma, separation) tuple
        legend: a lengend

    """
    width = (bins[1] - bins[0])
    X = bins[:-1] + (0.5*width)

    weights = np.sqrt(n)
    # weights = np.ones(n.shape[0])
    model = Model(fTwoPeak)
    sum = np.sum(n)/2.0
    params = Parameters()
    params.add('A1', value=sum)
    params.add('A2', value=sum)
    params.add('center', value=center)
    params.add('sigma', value=sigma, min=0.5*sigma, max=1.5*sigma)
    params.add('separation', value=separation, min=0.5*separation, max=1.5*separation)
    result = model.fit(n, params, x=X, weights=weights)

    center = result.best_values['center']
    sigma = result.best_values['sigma']
    separation = result.best_values['separation']
    legend = r'center=%.3f $\sigma$=%.1f sep=%.3f' % (center, sigma, separation)
    return result, (center, sigma, separation), legend


def fit_landau_model(mu, sigma, n, bins, debug=None):
    """Fit a Landau model.

    Fit Parameters:
        mu: estimated mean
        sigma: estimated sigma
        n: histogram bin  contents
        bins: bin limits

    Returns
    -------
        result: The fit result object
        out: a (mode, std) tuple
        legend: a lengend

    """
    width = (bins[1] - bins[0])
    X = bins[:-1] + (0.5*width)

    model = Model(fLandau)
    params = model.make_params(norm=np.sum(n), mode=mu, width=sigma)
    result = model.fit(n, params, x=X, weights=np.sqrt(n))

    mode = result.best_values['mode']
    width = result.best_values['width']
    legend = r'mode=%.3f $\sigma$=%.1f' % (mode, width)
    return result, (mode, width), legend


def fit_gaussian(n, bins, center, width=5.0, amplitude=1):
    """Fit a gaussion.

    Args:
    ----
        n: The bins
        bins: the bin edges
        center: The center (or mean) of the gaussian.
        width: the sigma estimate of the gaussion. Defaults to 5.0.
        amplitude: the estimae of the amplitude. Defaults to 1.

    Returns
    -------
        the fit result and a legend

    """
    step = 0.5 * (bins[1] - bins[0])
    X = bins[:-1] + step

    model = GaussianModel()
    params = model.make_params(amplitude=amplitude, center=center, sigma=width)
    result = model.fit(n, params, x=X)
    legend = r'$\mu$=%.3f $\sigma$=%.3f' % (result.best_values['center'], result.best_values['sigma'])
    return result, (result.best_values['center'], result.best_values['sigma']), legend


def fit_multi_gaus(hints, n, bins):
    """Fit a a number of gaussians as defined by the hints.

    Fit Parameters:
        hints: an array of (ampl, mean, sigma)
        n: histogram bin  contents
        bins: bin limits

    Returns
    -------
        result: The fit result object
        out: an array of (mean, std) tuples, one per peak
        legend: a lengend

    """
    width = (bins[1] - bins[0])
    X = bins[:-1] + (0.5*width)

    model = create_multi_peak(hints)
    # do the fit
    result = model.fit(n, x=X)
    legend = r""
    out = []
    for i in range(len(hints)):
        pref = "f{0}_".format(i)
        if i:
            legend += '\n'
        legend += r"$\mu$={:.3f} $\sigma$={:.3f}".format(result.best_values[pref + 'center'],
                                                         result.best_values[pref + 'sigma'])

        out.append((result.best_values[pref + 'center'], result.best_values[pref + 'sigma']))

    return result, out, legend


def fit_peak_model(n, bins, distance=10, debug=None):
    """Fits a multigaussian model from the number of peaks found.

    TODO: need to compute the distance from the noise and step

    Fit Parameters:
        n: histogram bin  contents
        bins: bin limits

    Returns
    -------
        result: The fit result object
        out: an array of (mean, std) tuples, one per peak
        legend: a lengend

    """
    width = (bins[1] - bins[0])
    ntot = np.sum(n)
    thrs = 0.01*ntot

    if debug:
        print("ntot:", ntot)
        print("thrs:", thrs)
        print("width:", width)

    peaks, prop = find_peaks(n, thrs, distance=distance)
    if debug:
        print("== Peaks")
        for peak, ampl in zip(peaks, prop['peak_heights']):
            print("\t height {:.1f} peak {:.1f} width {:.1f}".format(ampl, bins[peak], 3.46*width))

    hints = []
    for peak, ampl in zip(peaks, prop['peak_heights']):
        hints.append((ampl, bins[peak], 3.46 * width))

    return fit_multi_gaus(hints, n, bins)


def draw_best_fit(ax, result, bins, npts=100, legend=None, color="#fa6e1e"):
    """Draw the best fit.

    Args:
    ----
        ax: Axis to draw in.
        result: Result object from fit.
        bins: the histogram bins
        npts: number of points to evaluate fit function
        legend (optional): A legend for the plot. Defaults to None.
        color (optionan): color of fit curve

    """
    if npts < len(bins)+1:
        npts = len(bins)

    X = np.linspace(bins[0], bins[:-1], num=npts)
    Y = result.eval(param=result.params, x=X)
    ax.plot(X, Y, color=color)
    if legend is not None:
        ax.legend([legend], loc=1)


if __name__ == "__main__":
    nevts = 10000
    fig, ax = plt.subplots(nrows=1, ncols=2)

    # Double peak gauss
    peak1 = np.random.default_rng().normal(10.0, 0.5, nevts)
    values = np.append(peak1, np.random.default_rng().normal(7.5, 0.75, int(0.75*nevts)))

    count, bins, ignored = ax[0].hist(values, 50)
    result, out, legend = fit_peak_model(count, bins, debug=True)
    ax[0].legend([legend], loc=1)
    draw_best_fit(ax[0], result, bins)

    # The landau distributions
    r = 2.5*landau.rvs(size=nevts) + 10.0
    count, bins, ignored = ax[1].hist(r, 50)
    result, out, legend = fit_landau_model(np.mean(r), np.std(r), count, bins)
    ax[1].legend([legend], loc=1)
    draw_best_fit(ax[1], result, bins)

    plt.show()



from maddaq.MadDAQData import VarTypes, MadDAQData
from maddaq.MadDAQModule import MadDAQModule, ModuleDataIterator
from maddaq.ScanManager import ScanManager, VarDefinition, VarValue, ScanPoint
from maddaq.Progress import ShowProgress
from maddaq.GTimer import GTimer

from maddaq.cmmds.analyze_data import analyze_data
from maddaq.cmmds.show_data import show_data
from maddaq.cmmds.getSpectrum import getSpectrum
from maddaq.cmmds.file_info import getFileInfo

__version__ = "0.7.14"



#!/usr/bin/env python3
"""Python script to show the data in a file."""

import sys
from pathlib import Path
from maddaq import MadDAQData


def getFileInfo():
    """Main entry."""
    try:
        # We open here the file with MadDAQData
        ifile = Path(sys.argv[1])
        maddaq = MadDAQData(ifile)

        print("\n++ {}".format(ifile))
        maddaq.show_info(True)

        for m in maddaq.modules.values():
            m.save_pedestals("module_{}.ped".format(m.id))

    except KeyError:
        print("I need an input file")


if __name__ == "__main__":
    getFileInfo()


#!/usr/bin/env python3
"""ScanManager is a class that helps navigating the scan data."""
import sys


class VarDefinition(object):
    """Get the information relative to a variable in the scan."""

    def __init__(self, data):
        """Initialization."""
        self.type = data[0]
        self.npoints = data[1]
        self.first = data[2]
        self.last = data[3]

    def __str__(self):
        """String representation."""
        return "Type {}, npts {} from {} to {}".format(self.type,
                                                       self.npoints,
                                                       self.first,
                                                       self.last)


class VarValue(object):
    """The value of a scan variable."""

    def __init__(self, var_type, var_value):
        """Initialization."""
        self.type = var_type
        self.value = var_value

    def __str__(self):
        """String representation."""
        return "({tp}, {val})".format(tp=self.type,
                                      val=self.value)

    def __repr__(self):
        """REpresentation."""
        return self.__str__()


class ScanPoint(object):
    """A point in the scan."""

    def __init__(self, start, end, values):
        """Initialization."""
        self.start = start
        self.end = end
        self.values = values

    def __str__(self):
        """String represetnation."""
        if self.end != sys.float_info.max:
            return "{start:6.1f}:{end:6.1f} {values}".format(start=self.start,
                                                             end=self.end,
                                                             values=self.values)
        else:
            return "{start:6.1f}:  ---- {values}".format(start=self.start,
                                                         values=self.values)


class ScanIterator(object):
    """Itarator of a scan data file."""

    def __init__(self, scm):
        """Intialization.

        Args:
        ----
            scm (ScanManager): The ScanManager.

        """
        self.scm = scm
        self.i = 0

    def __iter__(self):
        """Return iterator  object, i.e, itself."""
        self.i = 0
        return self

    def __next__(self):
        """Next ScanPoint."""
        if self.i < self.scm.npoints:
            start = self.scm.start[self.i]
            end = self.scm.end[self.i]
            if end == start:
                # this is the last point
                end = sys.float_info.max

            values = {}
            jv = self.i*self.scm.nvar
            for val in self.scm.values[jv:jv+self.scm.nvar]:
                values[val[0]] = val[1]

            self.i += 1
            return ScanPoint(start, end, values)

        else:
            raise StopIteration()


class ScanManager(object):
    """Looks at the scan information on the file."""

    def __init__(self, maddaq):
        """We pass as argument a MadDAQData object."""
        try:
            self.nevts = maddaq.F["/scan/def/nevt"][0]
            self.nvar = maddaq.F["/scan/def/nvar"][0]
            self.variables = []
            for i in range(0, self.nvar):
                var = VarDefinition(maddaq.F["/scan/def/variables"][i])
                self.variables.append(var)

            self.start = maddaq.F["/scan/points/start"]
            self.end = maddaq.F["/scan/points/end"]
            self.values = maddaq.F["/scan/points/values"]
            self.npoints = int(self.values.shape/self.nvar)

        except KeyError:
            return

    def show_info(self):
        """Show Scan info."""
        print("Scan Information:")
        print("-----------------")
        print("Number of events {}".format(self.nevts))
        print("Number of variables {}.".format(self.nvar))
        print("Number of ScanPoints {}".format(self.npoints))
        for v in self.variables:
            print("+- {}".format(v))

    def iter(self):
        """Returns an iterator."""
        return ScanIterator(self)



#!/usr/bin/env python3
""" Test time in data
"""
import sys
import math
from pathlib import Path
from maddaq import MadDAQData, MadDAQModule
import numpy as np
import collections


class TimeCluster(object):
    """Cluster events by time."""

    def __init__(self, cut=1.0) -> None:
        """Initialization.

        Args:
            cut: Time cut. Defaults to 1.

        """
        self.cut = cut
        self.mean = 0.0
        self.n = 0.0
        self.ids = {}

    def check_event(self, mid, tdc):
        """Check if point is compatible with cluster.

        Args:
            mid: Module id
            tdc: Module tdc

        """
        if self.n == 0.0:
            self.add(mid, tdc)
            return True
        else:
            if abs(tdc - self.mean) < self.cut:
                self.add(mid, tdc)
                return True
            else:
                return False

    def add(self, mid, tdc):
        """Add a new point.

        Args:
            mid: Module id
            tdc: Module tdc

        """
        N = self.n + 1
        self.mean = (self.mean * self.n + tdc)/N
        self.n += 1.
        self.ids[mid] = tdc

    def std(self):
        return np.std(list(self.ids.values()))

    def get_boards(self):
        boards = {}
        for mid, tdc in self.ids.items():
            iboard = int(mid/10)
            boards.setdefault(iboard, []).append(tdc)

        return boards

    def reset(self):
        """Resets all counters."""
        self.n = 0.0
        self.mean = 0.0
        self.x2 = 0.0
        self.ids = {}


def get_period(maddaq):
    """Measure period or period.

    WE do this with an external trigger produced by a pulse generator.
    Same pulse for all the MB.

    Args:
    ----
        maddaq: The maddaq object.

    Return
    ------
        period (dict), std(dict) - period and deviation per MB id.

    """
    avg = {}
    last_time = {}
    for ievt, evt in enumerate(maddaq):
        evt_time = int(evt.time)
        mid = evt.mod_id
        board_id = int(int(mid)/10)
        ltim = last_time.get(mid, -1)
        if mid not in avg:
            avg[mid] = []

        dt, last_time[mid] = MadDAQModule.get_delta_time(evt_time, ltim)
        if dt != 0:
            avg[mid].append(dt)

        if ievt > 200:
            break

    period = {}
    rms = {}
    for key, val in avg.items():
        period[key] = np.mean(val)
        rms[key] = np.std(val)

    return period, rms


def main(fname, options):
    """fname is the input filename"""
    inp_file_path = Path(fname)
    if not inp_file_path.exists():
        print("Input file ", inp_file_path, " does not exist")
        return

    # open the maddaq file
    maddaq = MadDAQData(fname)
    maddaq.show_info()

    #  Get the ext. trigger period and guess number of boards.
    period, period_std = get_period(maddaq)
    tmp = {}
    for k in period_std.keys():
        tmp.setdefault(int(k/10), 0)

    nboard = len(tmp)
    nmodule = len(period)
    mean_period = np.mean(list(period.values()))
    std_period = np.mean(list(period_std.values()))

    print("nboard {}".format(nboard))
    print("nmodules {}".format(nmodule))
    print("avg period {:.3f}".format(mean_period))
    print("period std: {:.4f}".format(std_period))

    last_time = {}
    last_tdc = {}
    for m in list(maddaq.modules.values()):
        last_tdc[m.id] = 0.0

    # Get events sorted by TDC
    maddaq.iter_key = 'tdc'

    # Title of output
    print("   Row   mid  evtcnt          time     deltaT   deltaP")

    delta_board = []
    llog = collections.deque(maxlen=10)
    CL = TimeCluster(0.1*mean_period)
    Nevts = 0
    Nerr = 0
    N_spread = 0
    N_dist = 0
    N_both = 0
    for ievt, evt in enumerate(maddaq):
        evt_time = int(evt.time)
        mid = evt.mod_id
        board_id = int(int(mid)/10)
        ltim = last_time.get(mid, -1)

        dt, last_time[mid] = MadDAQModule.get_delta_time(evt_time, ltim)

        if options.is_pulse and dt > 0:
            offs = abs(dt - period[mid])/period[mid]
            if offs > 0.05:
                print("{:>6d} {:3d} Divergence {:>10.3f} expected {:>10.3f} [{:12X}]".format(ievt,
                                                                                             mid,
                                                                                             dt,
                                                                                             period[mid],
                                                                                             evt_time))

        # Compera with the last time for this module
        delta = evt.tdc - last_tdc[mid]                         # delta time
        delta_period = abs(delta - period[mid])     # faction of period

        # Store this tdc for next comparison.
        last_tdc[mid] = evt.tdc
        if dt == 0:
            delta_period = 0

        try:
            out = "{:>6d}   {:3d}  {:>6n} {:>12.3f} {:>10.3f}   {:.4f} [{:12X}]".format(
                ievt, mid, evt.evtcnt, evt.tdc, dt, delta_period, evt_time)

        except Exception:
            out = "{:>6d}   {:3d}  {:>6n} {:>12.3f} {:>10.3f}   {:.4f} [{:12X}]".format(
                ievt, mid, evt.evtcnt, evt.tdc, dt, 0.0, evt_time)

        llog.appendleft(out)

        # Analysis
        if not CL.check_event(mid, evt.tdc):
            if ievt/nmodule < options.show:
                print('-')
                
            # seems we are at the end of a trigger.
            Nevts += 1
            nerr = 0
            do_print = False
            msg = []
            # Complain if the std of times within cluster is large.
            if CL.std() > 1:
                N_spread += 1
                do_print = True
                nerr += 1
                msg.append("Event time spread larger than limit.")

            boards = CL.get_boards()
            btime = []
            for ib, val in boards.items():
                stdev = np.std(val)
                avg = np.mean(val)
                btime.append(avg)

            # Check the tome distance between boards
            if len(btime) > 1:
                dt_board = []
                for i in range(nboard-1):
                    for j in range(i+1, nboard):
                        dst = btime[i] - btime[j]
                        dt_board.append(abs(dst))

                mxdist = np.amax(dt_board)
                if mxdist != 0:
                    delta_board.append(mxdist)

                if mxdist > 0.1:
                    nerr += 1
                    N_dist += 1
                    do_print = True
                    msg.append("Max distance between boards: {:.6f}".format(mxdist))

            if do_print:
                Nerr += 1
                if nerr > 1:
                    N_both += 1

                print("========")
                for m in msg:
                    print("+ {}".format(m))
                print("")
                print("{} CL mean: {:.3f} std: {:.4f}".format(ievt, CL.mean, CL.std()))
                for ib, val in boards.items():
                    stdev = np.std(val)
                    avg = np.mean(val)
                    print("... board {}. mean {:.3f} std {:.4f}".format(ib, avg, stdev))

                print("... Max. Delta board: {:.4f}\n".format(mxdist))
                print("")
                print("   Row   mid  evtcnt         time     deltaT   deltaP      Register")
                print("---------------------------------------------------------------------")
                for x in reversed(llog):
                    print(x)

            # Reset the cluster and add the outsider to start again
            CL.reset()
            CL.add(mid, evt.tdc)
            
        if ievt/nmodule < options.show:
            print(out)


    print("Number of triggers: {}".format(Nevts))
    print("Number of errors {}".format(Nerr))
    print("Number of spread errors {}".format(N_spread))
    print("Number board dist {}".format(N_dist))
    print("Both errors {}".format(N_both))
    print("Average delta_board: {:.6f} - std {:.6f}".format(np.mean(delta_board), np.std(delta_board)))


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('files', nargs='*', help="Input files")
    parser.add_argument("--pulse", dest='is_pulse', action="store_true", default=True)
    parser.add_argument("--source", dest='is_pulse', action="store_false", default=True)
    parser.add_argument("--show_nevts", dest="show", type=int, default=50, help="Number of events to show.")
    options = parser.parse_args()
    if len(options.files) == 0:
        print(sys.argv[0])
        print("I need an input file")
        sys.exit()
    
    main(options.files[0], options)



#!/usr/bin/env python3
"""MadDAQData is a class that helps navigating the data."""

import os
import h5py
import heapq
import datetime

try:
    from MadDAQModule import MadDAQModule, ModuleDataIterator
except ModuleNotFoundError:
    from maddaq.MadDAQModule import MadDAQModule, ModuleDataIterator

try:
    from ScanManager import ScanManager
except ModuleNotFoundError:
    from maddaq.ScanManager import ScanManager

from enum import Enum
import operator


class VarTypes(Enum):
    """Types of Scan variables."""

    (CHANNEL, PULSE, THRESHOLD, TRIM,
     XPOSITION, YPOSITION, ZPOSITION,
     ANGLE1, ANGLE2, ANGLE3,
     USER1, USER2, USER3, USER4,
     THRS_TRIM, CAL_DAC) = range(16)


def get_run_info_time(T):
    """Convert the time struct in the data to a datetime."""
    return datetime.datetime(1900+T[5], T[4]+1, T[3], T[2], T[1], T[0])


def merge_iterators(iterables):
    """Merge iterables."""
    entries = {}  # Map from id to [front value, id, iterator].
    for id, it in enumerate(map(iter, iterables)):
        try:
            entries[id] = [next(it), id, it]
        except StopIteration:
            pass

    while entries:
        for id, entry in entries.items():
            value, _, it = entry
            yield value

            try:
                entry[0] = next(it)
            except StopIteration:
                del entries[id]
                break


class MadDAQData(object):
    """MadDAQData is the portal to the MadDAQ data.

    It maintains a list of modules and provides an iterator for the data.

    """

    def __init__(self, fname):
        """Open the file."""
        # Open the file
        if not os.path.exists(fname):
            raise Warning("Input file", fname, "does not exist")

        self.F = h5py.File(fname, 'r')

        # Find start and stop run records
        rr = self.F["/header/run_records"]
        self.run_type = rr[0]["run_type"]

        # Key for iterator
        self.iter_key = 'evt_time'

        # Find modules
        self.nevts = 0
        self.modules = {}
        for m in self.F["/header/modules"]:
            id = m & 0xff
            M = MadDAQModule(id, self.F)
            self.modules[id] = M
            self.nevts += M.data.shape[0]

        # The ScanManager. It will be None if not a ScanRun
        self.scm = self.scan_manager()

    def show_info(self, show_modules=False):
        """Show information about the current file.

        Args:
        ----
            show_modules: If True, print list of modules.

        """
        rr = self.F["/header/run_records"]
        t0 = get_run_info_time(rr[0]['time'])
        t1 = get_run_info_time(rr[1]['time'])
        print("Run: {} of type {}".format(rr[0]['run_number'],
                                          rr[0]['run_type']))
        print("     no. of events {}".format(rr[1]['nevts']))
        print("     Started: {}".format(t0.ctime()))
        print("     Finished: {}".format(t1.ctime()))
        print("     Duration: {}".format(t1-t0))

        if show_modules:
            print("\nNumber of modules: ", len(self.modules))
            for m in self.modules.values():
                print("+ Module: ", m.id,
                      " n. evts", m.data.shape[0],
                      " fw: %d.%d" % ((m.firmware & 0xff00) >> 8, m.firmware & 0xff))

    def get_duration(self):
        """Return the run duration in seconds."""
        rr = self.F["/header/run_records"]
        t0 = get_run_info_time(rr[0]['time'])
        t1 = get_run_info_time(rr[1]['time'])
        return (t1-t0).total_seconds()

    def has_scan(self):
        """Tell if the file has scan data."""
        try:
            self.F["/scan"]
            return True
        except KeyError:
            return False

    def scan_manager(self):
        """Return a ScanManager if there is Scan data or None otherwise."""
        if self.has_scan():
            return ScanManager(self)

        else:
            return None

    def scan_iter(self):
        """Return a scan iterator."""
        if self.scm:
            return self.scm.iter()

        else:
            return None

    def __iter__(self):
        """Iterate over all the modules.

        We sort them by time. We assume that data from each module
        is already sorted.
        TODO: find a way to pass another key function
        """
        return heapq.merge(*iter(list(self.modules.values())),
                           key=operator.attrgetter(self.iter_key))

    @staticmethod
    def check_module_list(mod_ids):
        """Check the list of module identifiers."""
        mids = None
        try:
            mids = [x for x in mod_ids]
        except TypeError:
            if mod_ids is not None:
                mids = [mod_ids]

        return mids

    def create_module_iterator(self, md, start=None, stop=None):
        """Create a module iterator."""
        return ModuleDataIterator(md, start, stop)

    def create_iterator_at_time(self, T, mod_ids=None):
        """Create an iterator that starst iterating at given daq_time.

        Args:
        ----
            T: daq_time to start
            mod_ids: list of valid module ID. None for all modules.

        """
        mod_iter = []
        mids = self.check_module_lists(mod_ids)

        for md in self.modules.values():
            if mids and md.id not in mids:
                continue

            ievt = md.find_time(T)
            mod_iter.append(ModuleDataIterator(md, start=ievt))

        if len(mod_iter):
            return merge_iterators(mod_iter)
        else:
            raise StopIteration

    def create_iterator_at_event(self, ievts, mod_ids=None):
        """Creatte an iterator that starts at a given event.

        Args:
        ----
            ievts: can be a single event or a tuple withe starting event
                   of each module
            mod_ids: list of valid module ID. None for all modules.

        """
        nm = len(self.modules)
        mids = self.check_module_lists(mod_ids)

        try:
            ninp = len(ievts)
            if ninp == 1:
                evt_list = [ievts[0] for x in range(0, nm)]

            elif ninp != nm:
                print("I need either a single event number or a list with one per module")
                return None

            else:
                evt_list = ievts

        except TypeError:
            evt_list = [x for x in range(0, nm)]

        mod_iter = []
        for iev, md in zip(evt_list, self.modules.values()):
            if mids and md.id not in mids:
                continue

            mod_iter.append(ModuleDataIterator(md, start=iev))

        if len(mod_iter):
            return merge_iterators(mod_iter)
        else:
            raise StopIteration

    def create_iterator_at_scanpoint(self, ipoint, mod_ids=None):
        """Create an iterator that starts at the given scan point.

        Args:
        ----
            ipoint: scan point to start
            mod_ids: list of valid module ID. None for all modules.

        """
        if not self.has_scan():
            raise StopIteration

        mids = self.check_module_list(mod_ids)

        if ipoint < 0 or ipoint >= self.scm.npoints:
            raise StopIteration

        mod_iter = []
        for md in self.modules.values():
            if mids and md.id not in mids:
                continue

            mevt = self.F["/scan/points/{}".format(md.id)][ipoint]
            if ipoint < self.scm.npoints-1:
                endp = self.F["/scan/points/{}".format(md.id)][ipoint+1]
                mod_iter.append(ModuleDataIterator(md, start=int(mevt), stop=int(endp)))
            else:
                mod_iter.append(ModuleDataIterator(md, start=int(mevt)))

        if len(mod_iter):
            return merge_iterators(mod_iter)
        else:
            raise StopIteration



#!/usr/bin/env python3
"""Example of data analysis with python readin hdf5 data."""
import os
import sys
import traceback

from maddaq.cmmds import fit_utils
import matplotlib.pyplot as plt
import numpy as np
from maddaq import MadDAQData
from maddaq import Progress
from scipy import stats


def read_data(maddaq, options, evt_E=None, evt_CH=None):
    """Iterates on all the events."""
    if options.nevts is None:
        max_evts = maddaq.nevts
    else:
        max_evts = options.nevts

    prg = Progress.ShowProgress(max_evts)
    prg.start()

    # now we iterate on all the data
    # for that we use maddaq as iterable
    if evt_E is None:
        evt_E = {}

    if evt_CH is None:
        evt_CH = {}

    print("\nIterating over all data in the file")
    try:
        for ievt, evt in enumerate(maddaq):
            if ievt == 0:
                continue

            if ievt > max_evts:
                break

            mid = evt.mod_id
            md = maddaq.modules[mid]
            data = md.process_event(evt)
            # get the array to store the event energy
            if data is not None:
                e_list = evt_E.setdefault(mid, [])
                e_ch = evt_CH.setdefault(mid, [])
                for channel, E in data:
                    if E > options.thrs:
                        e_list.append(E)
                        e_ch.append(channel)

            # monitor progress
            prg.increase()
            if prg.timer.mark() > 0.1:
                prg.timer.set_mark()
                prg.show()

    except RuntimeError:
        print("\n\nSomething very bad happened...")
        print('-'*60)
        traceback.print_exc(file=sys.stdout)
        print('-'*60)

    except KeyboardInterrupt:
        print("\nArrrrggggg !!!!")
        raise KeyboardInterrupt

    print("\nDone: ", ievt)
    return evt_E, evt_CH


def do_analyze_data(file_list, options):
    """Main routine."""

    evt_E = {}
    evt_CH = {}
    for fname in file_list:
        # Open the file
        if not os.path.exists(fname):
            print("Input file", fname, "does not exist")
            continue

        print("\n+++Reading file {}".format(fname))

        # We open here the file
        maddaq = MadDAQData(fname)
        for m in list(maddaq.modules.values()):
            m.set_debug(options.debug)
            m.do_cluster = options.cluster
            m.polarity = options.polarity

        maddaq.show_info(True)
        # and read the data
        evt_E, evt_CH = read_data(maddaq, options,
                                  evt_E=evt_E,
                                  evt_CH=evt_CH)

    # Create the figure
    n_mod = len(evt_E)
    fig, ax = plt.subplots(nrows=2, ncols=n_mod)
    dax = {}   # Spectrum axes
    daxc = {}  # Hitmap axes
    try:
        for iax, k in enumerate(evt_E.keys()):
            dax[k] = ax[0, iax]
            daxc[k] = ax[1, iax]

    except Exception:
        key = list(evt_E.keys())[0]
        dax[key] = ax[0]
        daxc[key] = ax[1]

    print("")

    # Do the fit
    for mid, E in evt_E.items():
        #  bit of statistics. Find out the optimal binwidth based on average noise
        vmin = np.min(E)
        vmax = np.max(E)
        range = vmax - vmin
        noise = stats.mode(maddaq.modules[mid].noise, keepdims=True).mode[0]
        if noise < 5.0:
            noise = 5.0

        if options.nbin < 0:
            nbin = int(range/(1.0*noise)/5.0 + 0.5) * 5
        else:
            nbin = options.nbin

        # Create and draw the histogram
        ax = dax[mid]
        n, bins, patches = ax.hist(E, nbin, range=(vmin, vmax), facecolor='g', alpha=0.75)
        step = 0.5 * (bins[1] - bins[0])
        X = bins[:-1] + step

        # Do the fits
        if options.fit == "landau":
            mu = np.mean(E)
            sigma = np.std(E)
            result, out, legend = fit_utils.fit_landau_model(mu, sigma, n, bins, options.debug)

        elif options.fit == "gauss":
            result, out, legend = fit_utils.fit_peak_model(n, bins, debug=options.debug)

        if options.fit != "none":
            if options.debug:
                print(result.fit_report())

            fit_utils.draw_best_fit(ax, result, bins)
            ax.legend([legend], loc=1)

        ax.set_title('Module %d' % mid)

        # Hit map
        for mid, CH in evt_CH.items():
            ax = daxc[mid]
            m = maddaq.modules[mid]
            X = np.arange(m.ntot+1)
            n, bins, patches = ax.hist(CH, facecolor='g', bins=m.ntot, alpha=0.75)
            ax.set_xlabel('Channel')
            ax.set_ylabel('Occupancy')

    plt.show()


def analyze_data():
    """Main entry."""
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('files', nargs='+', help="Input files")
    parser.add_argument("--fit", choices=["none", "gauss", "landau"],
                        help="It will try a Landau-like fit to the data (default)",
                        default="none"
                        )
    parser.add_argument("--debug", action="store_true", help="Debug", default=False)
    parser.add_argument("--cluster", action="store_true",
                        help="Will try to find clusters in sparse",
                        default=False)
    parser.add_argument("--nevts", help="Read at most nevts events", default=None)
    parser.add_argument("--thrs", help="Signal Threshold", default=-9999, type=float)
    parser.add_argument("--polarity", help="Signal polarity", default=1)
    parser.add_argument("--nbin", default=-1, type=int, help="Number of bins in histogram.")

    options = parser.parse_args()

    if len(options.files) == 0:
        print("I need an input file")
        sys.exit()

    do_analyze_data(options.files, options)
    sys.exit()


if __name__ == "__main__":
    analyze_data()



#!/usr/bin/env python3
"""MadDAQ module data class."""
import numpy as np
from collections import namedtuple


def __dummy_cond__(xx):
    return True


class MadDAQModule(object):
    """Module data handling class."""

    SERIAL, SPARSE, SPARSE_ADJ = (1, 2, 4)

    def __init__(self, module, data_file, polarity=1.0, do_cluster=False):
        """INitialize.

        Args:
        ----
            module: The module daq ID.
            data_file: The hdf5 data file object.
            polarity (optional): Polarity of the signal. Defaults to 1.0.
            do_cluster (optional): If cluster analysis is done. Defaults to False.

        """
        path = "/modules/%s/data/maddaq" % module
        self.data = data_file[path]
        path = "/modules/%s/data/time" % module
        self.time = data_file[path]

        self.polarity = polarity
        self.do_cluster = do_cluster
        names = ['mod_id', 'evt_time', 'tdc']
        names.extend(self.data.dtype.names)
        self.type = namedtuple("MadDAQModuleData", names)

        # Get pedeestals
        self.P = data_file["/modules/%s/pedestal/pedestal" % module]
        self.N = data_file["/modules/%s/pedestal/noise" % module]
        self.Ptime = data_file["/modules/%s/pedestal/evt_time" % module]

        # This is the pedestal and noise values used by default
        self.pedestal = self.P[-1]
        self.noise = self.N[-1]
        self.ntot = self.pedestal.shape[0]
        self.full_noise = [[] for i in range(0, self.ntot)]

        # Get module configuration from header
        self.id = int(module)
        self.nchip = -1
        modules = data_file["/header/modules"]
        for m in modules:
            id = m & 0xff
            if id == int(module):
                self.nchip = (m >> 20) & 0xff
                break

        # Read configuration record
        config = data_file["/modules/%s/configuration" % module]
        self.firmware = config[0] | config[1] >> 16
        self.threshold = config[2]
        self.ro_mode = config[3]
        self.nadj = config[4]
        self.hold_delay = config[5]
        self.ctw = config[6]
        self.trg_type = config[7]
        self.nbias = config[8]
        self.mbias = config[9:]
        self.adjacents = [0]

        for ch in range(1, self.nadj+1):
            self.adjacents.extend((ch, -ch))

        self.adjacents = np.array(self.adjacents)
        self.debug = False
        self.evdisplay = None
        self.cnvs = None

        self.seed_cut = 5.0
        self.neigh_cut = 3.0

    @staticmethod
    def get_delta_time(evt_time, last_time):
        """Compute the TDC time.

        Args:
        ----
            evt_time: current TDC time
            last_time: last time

        Return:
        ------
            dt: time since last event
            last_time: last absolute time

        """
        if last_time > 0:
            if evt_time > last_time:
                dt = evt_time - last_time
            else:
                dt = evt_time + (0x40000000-last_time)
        else:
            dt = 0.0

        dt = (dt*25.0)/1000.0
        return dt, evt_time

    def is_serial(self):
        """Tell if this is 'serial' data."""
        return self.ro_mode == MadDAQModule.SERIAL

    def set_current_pedestal(self, which=-1):
        """Set the default pedestals from the file."""
        self.pedestal = self.P[which]
        self.noise = self.N[which]

    def set_pedestals(self, pedestals, noise):
        """Set pedestal and noise."""
        self.pedestal = pedestals
        self.noise = noise

    def get_pedestal(self, which=-1):
        """Return pedestal."""
        return self.P[which]

    def get_noise(self, which=-1):
        """Return noise."""
        return self.N[which]

    def save_pedestals(self, file_name, which=-1):
        """Saves pedestals and noise to file."""
        ped = self.get_pedestal(which)
        noise = self.get_noise(which)
        ofile = open(file_name, "wt")
        i = 0
        for P, N in zip(ped, noise):
            ofile.write("{}, {:.2f}, {:.2f}\n".format(i, P, N))
            i = i+1

        ofile.close()

    def set_debug(self, dbg):
        """Set debug flag."""
        self.debug = dbg
        self.evdisplay = None

    def find_clusters(self, data, sn, hint=None, seed_cut=5.0, neigh_cut=3.0):
        """Do cluster analysis."""
        out = []

        #
        # find the clusters
        #
        used = [False for x in range(0, self.ntot)]

        # Get the list of channels with signal over noise > 5
        if hint is None:
            channels = np.nonzero(sn > seed_cut)[0]

        else:
            channels = [int(hint)]

        for ch in channels:
            if used[ch]:
                continue

            clst = [ch]
            used[ch] = True

            j = ch-1
            while True:
                if j < 0:
                    break

                if sn[j] > neigh_cut and not used[j]:
                    clst.append(j)
                    used[j] = True
                    j = j - 1
                else:
                    break

            j = ch + 1
            while True:
                if j > self.ntot - 1:
                    break

                if sn[j] > neigh_cut and not used[j]:
                    clst.append(j)
                    used[j] = True
                    j = j + 1
                else:
                    break

            Eclst = 0.0
            chan = 0.0
            for i in clst:
                if data[i] > 0.0:
                    Eclst += data[i]
                    chan += i*data[i]

            if Eclst > 0:
                chan /= Eclst
                out.append((chan, Eclst))

        return out

    def analyze_serial(self, evt):
        """Analysis of data in serial."""
        # subtract pedestals
        try:
            data = evt.data - self.pedestal
        except ValueError:
            # Sometimes it says we are serial, but we aren't
            return []

        if self.polarity < 0.0:
            data *= -1.0

        # compute signal over noise
        sn = data/self.noise

        # compute common mode and remove it
        # TODO: This is better done on a chip by chip basis
        cmmd = np.mean(data[np.nonzero(abs(sn) < 5.0)])
        data -= cmmd

        #
        # find the clusters
        #
        out = self.find_clusters(data, sn, seed_cut=self.seed_cut, neigh_cut=self.neigh_cut)

        return out

    def analyze_sparse(self, evt):
        """Analysis of sparse + adjacent data."""
        #
        # TODO: think something smarter than rejecting
        #       events with naighbours outside the range
        #
        the_chan = evt.nchan
        try:
            if evt.chan < self.nadj or evt.chan + self.nadj >= self.ntot:
                val = evt.data[0] - self.pedestal[evt.chan]
                return [(evt.chan, val)]
        except Exception as w:
            print("chan {} nadj {} - {}".format(evt.chan, self.nadj, str(w)))
            return []

        if evt.chan < 0 or evt.chan > self.ntot:
            print("chan {} nadj {}".format(evt.chan, self.nadj))
            return []

        # Get the indices for the pedestals ans subtract pedestals
        indx = [indx + the_chan for indx in self.adjacents]
        data = evt.data - self.pedestal[indx]
        if self.polarity < 0.0:
            data *= -1.0

        #
        # Subtract Common mode
        # Ideally this has to be done on a chip by chip basis
        #
        if self.nadj > 3:
            cmmd = np.mean(data[1:])
            data -= cmmd
        else:
            cmmd = 0.0

        out = []
        if evt.romode == MadDAQModule.SPARSE_ADJ and self.do_cluster:
            vsn = data/self.noise[indx]
            E = 0.0
            X = 0.0
            nstrip = 0
            i = 0
            while i < len(data):
                if i == 0:
                    val = data[i]
                    E += val
                    X += self.adjacents[i]*val
                    nstrip += val
                    i += 1
                else:
                    ngood = 0
                    for j in (i, i+1):
                        if vsn[j] > 5.0:
                            val = data[j]
                            E += val
                            X += self.adjacents[j]*val
                            nstrip += val
                            ngood += 1

                    i += 2
                    if ngood == 0:
                        break

            if nstrip > 0:
                X /= nstrip
                out.append((evt.chan+X, E))

        else:
            out.append((evt.chan, data[0]))

        return out

    def process_event(self, evt):
        """Very simple event processing."""
        if evt.romode == MadDAQModule.SERIAL:
            return self.analyze_serial(evt)

        elif evt.romode == MadDAQModule.SPARSE_ADJ:
            return self.analyze_sparse(evt)

        elif evt.romode == MadDAQModule.SPARSE:
            print("SPARSE not implemented")
            return None

    def find_time(self, T):
        """Find the event with DAQ time closest to the given."""
        ntot = self.time.shape[0]
        aa = 0
        bb = ntot-1
        faa = self.time[aa]
        fbb = self.time[bb]
        last_indx = -1
        while True:
            indx = aa+int((bb-aa)/2)
            # print("[{} {}] f(a) {}".format(aa, bb, faa))
            if indx == aa:
                return aa

            val = self.time[indx]

            if val > T:
                bb = indx
                fbb = val

            elif val < T:
                aa = indx
                faa = val

            if abs(aa-bb) < 1:
                return aa

    def __iter__(self):
        """Return the iterator."""
        # return self.navigate()
        return ModuleDataIterator(self)

    def navigate(self, start=None, stop=None, condition=__dummy_cond__):
        """The actual iterator routine..

        If start and stop are the first and last event we want read
        condition is a boolean function that receives the module data.
        navigate will only return the events for which condition
        returns true.
        """
        nevts = self.data.shape[0]

        # chunk size os the same for data and time
        chunk_size = 10*self.data.chunks[0]

        if start is None:
            start = 0

        if stop is None:
            stop = nevts

        if stop > nevts:
            stop = nevts

        first_chunk = start/chunk_size
        last_chunk = stop/chunk_size + 1
        current_chunk = first_chunk

        offs = 0
        for indx in range(start, stop):
            if indx % chunk_size == 0:
                TS = self.time[indx:indx+chunk_size]
                DS = self.data[indx:indx+chunk_size]
                offs = indx

            values = [self.id, TS[indx-offs]]
            values.extend(DS[indx-offs])
            obj = self.type._make(values)
            if condition(obj):
                yield obj


class ModuleDataIterator(object):
    """Iterator for module data."""

    def __init__(self, module, start=None, stop=None, condition=__dummy_cond__):
        """The actual iterator.

        start and stop are the first and last event we want to read
        condition is a boolean function that receives the module data.
        The iterator will only return the events for which condition
        returns true.
        """
        self.M = module
        self.start = start
        self.stop = stop
        self.condition = condition
        self.nevts = self.M.data.shape[0]
        self.chunk_size = 20*self.M.data.chunks[0]
        self.first_chunk = None
        self.last_chunk = None
        self.current_chunk = None

        self.TS = None
        self.DS = None
        self.first_in_chunck = 0
        self.indx = 0

        self.init_iter()

    def init_iter(self):
        """Inititalize the iterator."""
        if self.indx > 0:
            return

        if self.start is None:
            self.start = 0

        if self.stop is None or self.stop > self.nevts:
            self.stop = self.nevts

        self.first_chunk = int(self.start/self.chunk_size)
        self.last_chunk = int(self.stop/self.chunk_size) + 1
        self.current_chunk = self.first_chunk
        self.end_of_chunk = (self.current_chunk + 1) * self.chunk_size
        if self.end_of_chunk > self.stop:
            self.end_of_chunk = self.stop

        # offs is the beginning of the chuck
        self.first_in_chunck = int(self.current_chunk*self.chunk_size)
        self.indx = self.start
        self.TS = None
        self.DS = None

        self.last_time = -1
        self.the_time = 0.0

    def __iter__(self):
        """Return iterator."""
        self.init_iter()
        return self

    def __next__(self):
        """Iterator ext method."""
        if self.indx >= self.stop:
            raise StopIteration

        else:
            while True:
                # index within chunck
                ii = self.indx - self.first_in_chunck

                if self.TS is None or ii % self.chunk_size == 0:
                    self.first_in_chunck = self.indx
                    if self.TS is not None:
                        self.current_chunk += 1
                        self.end_of_chunk = (self.current_chunk + 1) * self.chunk_size
                        if self.end_of_chunk > self.stop:
                            self.end_of_chunk = self.stop

                    self.TS = self.M.time[self.first_in_chunck:self.end_of_chunk]
                    self.DS = self.M.data[self.first_in_chunck:self.end_of_chunk]
                    ii = 0

                # The TDC
                dt, self.last_time = MadDAQModule.get_delta_time(int(self.DS[ii][2]), self.last_time)
                self.the_time += dt

                # Create the event object
                values = [self.M.id, self.TS[ii], self.the_time]
                values.extend(self.DS[ii])
                obj = self.M.type._make(values)

                self.indx += 1
                if self.indx >= self.stop:
                    # print("ultimo {}".format(self.M.id))
                    raise StopIteration

                if self.condition(obj):
                    return obj



#!/usr/bin/env python3
""" An object usefull to show the progress of a process.
"""
import sys
try:
    from GTimer import GTimer
except ImportError:
    from maddaq.GTimer import GTimer

def rate_units(r):
    units = "Hz"
    if r > 1.0e6:
        r /= 1.0e6
        units = "MHz"

    elif r > 1.0e3:
        r /= 1.0e3
        units = "kHz"

    elif r > 1.0:
        pass

    elif r > 1e-3:
        r /= 1.0e-3
        units = "mHz"

    else:
        r /= 1.0e-6
        units = "uHz"

    return r, units


def time_units(t):
    units = "s"
    if t > 86400.0:
        t /= 86400.0
        units = "d "

    elif t > 3600.0:
        t /= 3600.0
        units = "h "

    elif t > 60.0:
        t /= 60.0
        units = "m "

    elif t > 1.0:
        units = "s "

    elif t > 1.0e-3:
        t /= 1.0e-3
        units = "ms"

    elif t > 1.0e-6:
        t /= 1.0e-6
        units = "us"

    else:
        t /= 1.0e-9
        units = "ns"

    return t, units


class ShowProgress(object):
    """ This class shows the program status.
    """

    def __init__(self, max_val, width=40):
        self.width = width-2
        self.timer = GTimer()
        self.counter = 0.0
        self.max_val = float(max_val)
        self.prg = 0.0

    def start(self):
        self.counter = 0
        self.timer.start()

    def stop(self):
        self.timer.stop()

    def increase(self, val=1.0, show=False, interval=0.1):
        self.counter += val
        self.prg = self.counter/self.max_val
        if show:
            if self.timer.mark() > interval:
                self.timer.set_mark()
                self.show()

    def show(self):
        self.show_stat(self.prg)

    def show_stat(self, x):
        n21 = int(x*self.width)
        n22 = int(self.width-n21-1)

        c21 = n21*'='
        c22 = n22*' '
        if self.prg > 0.0:
            tt = self.timer()*(1.0/self.prg-1.0)
        else:
            tt = 0.0

        rate = self.counter/self.timer()
        rv, ru = rate_units(rate)
        te, teu = time_units(self.timer())
        tr, tru = time_units(tt)

#        ss = '\r[%s>%s] %5.1f%% %8d' % (c21 , c22, 100.*x, self.counter)
        ss = '\rElapsed %4.1f %s %5.1f %s [%s>%s] %5.1f%% ERT %5.1f %s' % (te, teu, rv, ru, c21, c22, 100.*x,  tr, tru)
        print(ss, end='')
        sys.stdout.flush()



#!/usr/bin/env python3
"""Example of data analysis with python reading hdf5 data.

The example only shows how to iterate on the data.
A more complete example can be found in analyze_data.
"""
import os
import sys
from maddaq import MadDAQData, ShowProgress


def read_data(fname):
    """Main entry."""
    # Check the file
    if not os.path.exists(fname):
        print("Input file", fname, "does not exist")
        return

    # We open here the file
    maddaq = MadDAQData(fname)

    # iterate over the modules and show some data from each of them
    # This shows how to create a module iterator.
    for m in maddaq.modules.values():
        print("Module %d: n. chips %d" % (m.id, m.nchip))
        for evt in maddaq.create_module_iterator(m, stop=10):
            print(evt.daq_time, evt.time)

    # Now print data from the first module
    # Get the first module
    the_module = list(maddaq.modules.values())[0]
    print("\nThis is module", the_module.id)
    
    # The module itself is an iterator.
    for i, evt in enumerate(the_module):
        print(evt.daq_time, evt.data)
        if i > 10:
            break

    # now we iterate on all the data in the file
    # for that we use maddaq as iterator
    prg = ShowProgress(maddaq.nevts)
    prg.start()
    print("\nIterating over all data in the file\nTime should be sorted")
    try:
        for i, evt in enumerate(maddaq):
            if i < 25:
                print(evt.mod_id, evt.daq_time)

            else:
                prg.increase()
                if prg.timer.mark() > 0.1:
                    prg.timer.set_mark()
                    prg.show()

    except KeyboardInterrupt:
        print("\nArrrrggggg !!!!")

    print("\n")


if __name__ == "__main__":
    try:
        fname = sys.argv[1]
    except IndexError:
        print("I need an input file")
        sys.exit()

    read_data(fname)
